use crate::se::WriteError;

pub trait GenericWriteBuffer: Sized {
	type Error;

	fn clear(&mut self);

	fn push_byte(&mut self, byte: u8) -> Result<(), Self::Error>;

	fn push_slice(&mut self, slice: &[u8]) -> Result<(), Self::Error>;
}

pub trait GenericReadBuffer: Sized {
	fn len(&self) -> usize;

	fn as_slice(&self) -> &[u8];

	#[cfg(any(feature = "std", feature = "alloc"))]
	fn to_vec(self) -> Vec<u8> {
		self.as_slice().to_vec()
	}
}

///////////////////////////////////////////////////////////////////////////////

#[derive(Debug)]
pub struct SliceBufferError;

impl WriteError for SliceBufferError {}

pub struct SliceBuffer<'a> {
	buf: &'a mut [u8],
	len: usize
}

impl<'a> SliceBuffer<'a> {
	pub fn new(buf: &'a mut [u8]) -> Self {
		Self { buf, len: 0 }
	}
}

impl<'a> GenericReadBuffer for SliceBuffer<'a> {
	#[inline]
	fn len(&self) -> usize {
		self.len
	}

	fn as_slice(&self) -> &[u8] {
		&self.buf[..self.len]
	}
}

impl<'a> GenericWriteBuffer for SliceBuffer<'a> {
	type Error = SliceBufferError;

	#[inline]
	fn clear(&mut self) {
		self.len = 0;
	}

	fn push_byte(&mut self, byte: u8) -> Result<(), Self::Error> {
		let new_len = self.len + 1;
		if new_len > self.buf.len() {
			Err(SliceBufferError)
		} else {
			self.buf[self.len] = byte;
			self.len = new_len;
			Ok(())
		}
	}

	fn push_slice(&mut self, slice: &[u8]) -> Result<(), Self::Error> {
		let new_len = self.len + slice.len();
		if new_len > self.buf.len() {
			Err(SliceBufferError)
		} else {
			self.buf[self.len..new_len].copy_from_slice(slice);
			self.len = new_len;
			Ok(())
		}
	}
}

///////////////////////////////////////////////////////////////////////////////

#[cfg(any(feature = "std", feature = "alloc"))]
impl GenericWriteBuffer for Vec<u8> {
	// TODO: Use never type
	type Error = ();

	#[inline]
	fn clear(&mut self) {
		self.clear();
	}

	#[inline]
	fn push_byte(&mut self, byte: u8) -> Result<(), Self::Error> {
		Ok(self.push(byte))
	}

	#[inline]
	fn push_slice(&mut self, slice: &[u8]) -> Result<(), Self::Error> {
		Ok(self.extend_from_slice(slice))
	}
}

#[cfg(any(feature = "std", feature = "alloc"))]
impl GenericReadBuffer for Vec<u8> {
	fn len(&self) -> usize {
		self.len()
	}

	fn as_slice(&self) -> &[u8] {
		self.as_slice()
	}

	fn to_vec(self) -> Vec<u8> {
		self
	}
}